from functools import reduce
import operator

from django.utils.encoding import force_text
from django.utils.http import urlencode


def querystring(data, exclude=(), **kwargs):
    """
    Returns a properly encoded querystring

    The supported arguments are as follows:

    - ``data`` should be a ``MultiValueDict`` instance (i.e. ``request.GET``)
    - ``exclude`` is a list of keys from ``data`` which should be skipped
    - Additional key-value pairs are accepted as keyword arguments

    Usage::

        next_page_url = querystring(
            request.GET,
            exclude=('page',),
            page=current + 1,
            )
    """
    items = reduce(operator.add, (
        list((k, v) for v in values)
        for k, values in data.lists() if k not in exclude
    ), [])

    for key, value in kwargs.items():
        items.append((key, force_text(value)))

    return urlencode(sorted(items))



"""
Authentication backend which preloads access and client models
==============================================================
"""

from django.contrib.auth.models import User

from towel.auth import ModelBackend as _ModelBackend
from towel.mt import access_model, client_model


class ModelBackend(_ModelBackend):
    """
    Custom authentication backend for towel-mt

    This authentication backend serves two purposes:

    1. Allowing email addresses as usernames (``authenticate``)
    2. Minimizing DB accesses by fetching additional information about the
       current user earlier (``get_user``)
    """

    def get_user(self, user_id):
        Access = access_model()
        Client = client_model()
        try:
            access = Access.objects.select_related(
                'user',
                Client.__name__.lower(),
            ).get(user=user_id)

            # Ensure reverse accesses do not needlessly query the DB again.
            # Maybe Django already does that for us already... whatever.
            setattr(access.user, User.access.cache_name, access)
            return access.user
        except Access.DoesNotExist:
            pass

        try:
            # Fall back to raw user access
            return User.objects.get(id=user_id)
        except User.DoesNotExist:
            pass

        return None



from django.conf.urls import url as _url
from django.core.urlresolvers import NoReverseMatch, reverse

from towel import resources
from towel.utils import app_model_label


class _MRUHelper(object):
    def __init__(self, viewname_pattern, kwargs):
        self.viewname_pattern = viewname_pattern
        self.kwargs = kwargs

    def __getitem__(self, item):
        return self.url(item)

    def url(self, item, *args, **kwargs):
        kw = self.kwargs
        if kwargs:
            kw = kw.copy()
            kw['kwargs'].update(kwargs)

        try:
            return reverse(self.viewname_pattern % item, **kw)
        except NoReverseMatch as e:
            try:
                return reverse(self.viewname_pattern % item)
            except NoReverseMatch:
                # Re-raise exception with kwargs; it's more informative
                raise e


def model_resource_urls(
        reverse_kwargs_fn=lambda object: {'pk': object.pk},
        default='detail'):
    """
    Usage::

        @model_resource_urls()
        class MyModel(models.Model):
            pass

        instance = MyModel.objects.get(...)
        instance.urls.url('detail') == instance.get_absolute_url()
    """
    def _dec(cls):
        class _descriptor(object):
            def __get__(self, obj, objtype=None):
                viewname_pattern = '%s_%s_%%s' % app_model_label(obj)
                kwargs = {'kwargs': reverse_kwargs_fn(obj)}
                helper = obj.__dict__['urls'] = _MRUHelper(
                    viewname_pattern, kwargs)
                return helper

        cls.urls = _descriptor()
        cls.get_absolute_url = lambda self: self.urls.url(default)
        return cls
    return _dec


def resource_url_fn(
        model,
        urlconf_detail_re=r'(?P<pk>\d+)',
        mixins=(),
        decorators=(),
        **kwargs):
    """
    Returns a helper function most useful to easily create URLconf entries
    for model resources.

    The list of decorators should be ordered from the outside to the inside,
    in the same order as you would write them when using the ``@decorator``
    syntax.

    Usage::

        project_url = resource_url_fn(
            Project,
            mixins=(ProjectViewMixin,),
            decorators=(login_required,),
            )
        urlpatterns = patterns(
            '',
            project_url('list', url=r'^$', paginate_by=50),
            project_url('detail', url=r'^(?P<pk>\d+)/$'),
            project_url('add', url=r^add/$'),
            project_url('edit'),
            project_url('delete'),
        )

        # the project URLs will be:
        # ^$
        # ^(?P<pk>\d+)/$
        # ^add/$'
        # ^(?P<pk>\d+)/edit/$
        # ^(?P<pk>\d+)/delete/$

    The returned helper function comes with ``mixins`` and ``decorators``
    arguments too. They default to the values passed into the
    ``resource_url_fn``. If you use those arguments, you have to pass the
    full list of mixins and/or decorators you need. You can pass an empty
    list if some view does not need any mixins and/or decorators.
    """

    global_mixins = mixins
    global_decorators = decorators

    default_view_classes = {
        'list': resources.ListView,
        'detail': resources.DetailView,
        'add': resources.AddView,
        'edit': resources.EditView,
        'delete': resources.DeleteView,
    }

    def _fn(
            name,
            _sentinel=None,
            view=None,
            url=None,
            mixins=None,
            decorators=None,
            **kw):

        if _sentinel is not None:
            raise TypeError('name is the only non-keyword')

        urlregex = (
            r'^%s/%s/$' % (urlconf_detail_re, name)
            if url is None else url
        )

        urlname = '%s_%s_%s' % (app_model_label(model) + (name,))

        mixins = global_mixins if mixins is None else mixins
        decorators = global_decorators if decorators is None else decorators

        kws = kwargs.copy()
        kws.update(kw)

        view = default_view_classes[name] if view is None else view
        view = type(view.__name__, mixins + (view,), {})
        view = view.as_view(model=model, **kws)

        for dec in reversed(decorators):
            view = dec(view)

        return _url(urlregex, view, name=urlname)

    return _fn



"""
Helper module to circumvent Django's (arguably) broken formset saving behavior
where models are directly deleted even if ``commit=False`` is passed to the
formset's ``save()`` method.

Usage::

    class SafeModel(deletion.Model):
        # fields etc.


    instance = SafeModel.objects.get(...)
    with deletion.protect():
        instance.delete() # Does nothing
    instance.delete() # Actually deletes the instance!


Saving formsets::

    with deletion.protect():
        objects = formset.save()

    for obj in formset.deleted_objects: # Django provides this attribute
        obj.delete() # Or do something else, like checking whether the instance
                     # should really be deleted


This is achieved by overriding the model's ``delete()`` method with a different
version which does nothing if protection is active. If you override the
deletion method for some reason too, you have to ensure that the threadlocal
state is respected too.
"""

from contextlib import contextmanager
from threading import local

from django.db import models


DEFAULT = None
PROTECT = 'protect'

_deletion = local()


def set_mode(mode):
    """
    Sets the protection mode. The argument should be one of:

    - ``deletion.DEFAULT``:
      Standard behavior, instances are deleted.
    - ``deletion.PROTECT``:
      ``delete()`` invocations on models inheriting ``deletion.Model`` are
      ignored.
    """
    _deletion.mode = mode


@contextmanager
def protect():
    """
    Wraps a code block with deletion protection

    Example::

        from towel import deletion

        instance = SafeModel.objects.get(...)
        with deletion.protect():
            # Does nothing
            instance.delete()

        # Actually deletes the instance
        instance.delete()
    """
    set_mode(PROTECT)
    yield
    set_mode(DEFAULT)


class Model(models.Model):
    """
    Safe model base class, inherit this model instead of the standard
    :py:class:`django.db.models.Model` if you want to take advantage of
    the :py:mod:`towel.deletion` module.
    """
    class Meta:
        abstract = True

    def delete(self, *args, **kwargs):
        """
        Deletion is skipped if inside a :py:func:`~towel.deletion.protect`
        block.
        """
        if getattr(_deletion, 'mode', None) == PROTECT:
            return
        super(Model, self).delete(*args, **kwargs)



"""
Making ``towel.api`` multitenancy-aware
=======================================

All you need is a view decorator handling the permissions and a resource
subclass which makes sure that data is only ever shown from one tenant.
"""

from functools import wraps

from django.http import HttpResponse
from django.utils.six.moves import http_client

from towel import api
from towel.utils import safe_queryset_and


def api_access(minimal):
    """
    Decorator which ensures that the current ``request.access`` model
    provides at least ``minimal`` access.
    """
    def _decorator(func):
        @wraps(func)
        def _fn(request, *args, **kwargs):
            if not request.access:
                return HttpResponse('No access',
                    status=http_client.UNAUTHORIZED)

            if request.access.access < minimal:
                return HttpResponse('Insufficient access',
                    status=http_client.UNAUTHORIZED)

            return func(request, *args, **kwargs)
        return _fn
    return _decorator


class Resource(api.Resource):
    """
    Resource subclass which automatically applies filtering by
    ``request.access`` to all querysets used.
    """
    def get_query_set(self):
        return safe_queryset_and(
            super(Resource, self).get_query_set(),
            self.model.objects.for_access(self.request.access),
        )



"""
``ModelView``
=============
==
As long as you use this class, everything should just work (tm).
"""

from towel import modelview as towel_modelview

import towel.mt
from towel.mt.forms import ModelForm


class ModelView(towel_modelview.ModelView):
    """
    This model view subclass ensures that all querysets are already
    restricted to the current client.

    Furthermore, it requires certain access levels when accessing its
    views; the required access defaults to ``access.MANAGER`` (the
    highest access level), but can be overridden by setting
    ``view_access`` or ``crud_access`` when instantiating the model
    view.
    """

    #: Default access level for all views
    view_access = None

    #: Default access level for CRUD views, falls back to ``view_access``
    #: if not explicitly set
    crud_access = None

    #: The editing form class, defaults to ``towel.mt.forms.ModelForm``
    #: instead of ``django.forms.ModelForm``
    form_class = ModelForm

    def view_decorator(self, func):
        return towel.mt._access_decorator(self.view_access)(func)

    def crud_view_decorator(self, func):
        return towel.mt._access_decorator(
            self.crud_access or self.view_access)(func)

    def get_query_set(self, request, *args, **kwargs):
        return self.model.objects.for_access(request.access)

    def get_form_instance(self, request, form_class, instance=None,
            change=None, **kwargs):
        args = self.extend_args_if_post(request, [])
        kwargs.update({
            'instance': instance,
            'request': request,  # towel.mt.forms needs that
        })

        return form_class(*args, **kwargs)



import json
import re

from django.utils.six.moves import http_client

from .serializers import Serializer


class RequestParser(object):
    """
    Parses the request body into a format independent of its content type.

    Does nothing for the following HTTP methods because they are not supposed
    to have a request body:

    - ``GET``
    - ``HEAD``
    - ``OPTIONS``
    - ``TRACE``
    - ``DELETE``

    Otherwise, the code tries determining a parser for the request. The
    following content types are supported:

    - ``application/x-www-form-urlencoded`` (the default)
    - ``multipart/form-data``
    - ``application/json``

    The two former content types are supported directly by Django, all
    capabilities and restrictions are inherited directly. When using JSON,
    file uploads are not supported.

    The parsed data is available as ``request.POST`` and ``request.FILES``.
    ``request.POST`` is used instead of something else even for ``PUT`` and
    ``PATCH`` requests (among others), because most code written for Django
    expects data to be provided under that name.
    """
    def parse(self, request):
        """
        Decides whether the request body should be parsed, and if yes, decides
        which parser to use. Returns a HTTP 415 Unsupported media type if the
        request isn't understood.
        """
        if request.method in ('GET', 'HEAD', 'OPTIONS', 'TRACE', 'DELETE'):
            # Fall back to standard handling which only does stuff when
            # method == 'POST' anyway, that is, not now.
            return

        content_type = request.META.get('CONTENT_TYPE',
            'application/x-www-form-urlencoded')

        handlers = {
            r'^application/x-www-form-urlencoded': self.parse_form,
            r'^multipart/form-data': self.parse_form,
            r'^application/json': self.parse_json,
        }

        for pattern, handler in handlers.items():
            if re.match(pattern, content_type):
                return handler(request)

        return Serializer().serialize(
            {
                'error': '%r is not supported' % content_type,
            },
            request=request,
            status=http_client.UNSUPPORTED_MEDIA_TYPE,
            output_format=request.GET.get('format'),
        )

    def parse_form(self, request):
        """
        Simply calls Django's own request parsing, but changes the request
        method to ``'POST'`` so that the request body parsing actually happens
        as it should. The value of ``request.method`` is restored when
        parsing ends.
        """
        method = request.method
        request.method = 'POST'
        request._load_post_and_files()
        request.method = method

    def parse_json(self, request):
        """
        Unserializes the JSON in the body of the request and saves the result
        as ``request.POST``.
        """
        request.POST = json.loads(request.body.decode('utf-8'))



from functools import wraps

from django.contrib.auth import authenticate
from django.http import HttpResponse
from django.utils.six.moves import http_client
from django.views.decorators.vary import vary_on_headers


def http_basic_auth(func):
    @wraps(func)
    @vary_on_headers('Authorization')
    def _decorator(request, *args, **kwargs):
        if 'HTTP_AUTHORIZATION' in request.META:
            meth, _, auth = request.META['HTTP_AUTHORIZATION'].partition(' ')
            if meth.lower() == 'basic':
                try:
                    auth = auth.strip().decode('base64')
                except Exception:  # binascii.Error, really.
                    return HttpResponse('Invalid authorization header',
                        status=http_client.BAD_REQUEST)

                username, sep, password = auth.partition(':')
                user = authenticate(username=username, password=password)
                if user:
                    request.user = user

        return func(request, *args, **kwargs)
    return _decorator



import itertools

from django import template


register = template.Library()


PATHS = [
    '_meta',
    'queryset.model._meta',
    'instance._meta',
    'model._meta',
]


def _resolve(instance, last_part):
    for path in PATHS:
        o = instance
        found = True
        for part in itertools.chain(path.split('.'), [last_part]):
            try:
                o = getattr(o, part)
            except AttributeError:
                found = False
                break

        if found:
            return o


@register.filter
def verbose_name(item):
    """
    Pass in anything and it tries hard to return its ``verbose_name``::

        {{ form|verbose_name }}
        {{ object|verbose_name }}
        {{ formset|verbose_name }}
        {{ object_list|verbose_name }}
    """
    return _resolve(item, 'verbose_name')


@register.filter
def verbose_name_plural(item):
    """
    Pass in anything and it tries hard to return its ``verbose_name_plural``::

        {{ form|verbose_name_plural }}
        {{ object|verbose_name_plural }}
        {{ formset|verbose_name_plural }}
        {{ object_list|verbose_name_plural }}
    """
    return _resolve(item, 'verbose_name_plural')



import itertools
import re

from django.core.urlresolvers import NoReverseMatch, reverse
from django.db.models.deletion import Collector


def related_classes(instance):
    """
    Return all classes which would be deleted if the passed instance
    were deleted too by employing the cascade machinery of Django
    itself. Does **not** return instances, only classes.

    Note! When using Django 1.5, autogenerated models (many to many through
    models) are returned too.
    """
    collector = Collector(using=instance._state.db)
    # We really do not want fast deletion, we absolutely need to know whether
    # there are related objects around!
    collector.can_fast_delete = lambda *args, **kwargs: False
    collector.collect([instance])

    # Save collected objects for later referencing (well yes, it does return
    # instances but we don't have to tell anybody :-)
    instance._collected_objects = collector.data

    return collector.data.keys()


def safe_queryset_and(head, *tail):
    """
    Safe AND-ing of querysets. If one of both queries has its
    DISTINCT flag set, sets distinct on both querysets. Also takes extra
    care to preserve the result of the following queryset methods:

    * ``reverse()``
    * ``transform()``
    * ``select_related()``
    * ``prefetch_related()``
    """

    def _merge(qs1, qs2):
        if qs1.query.distinct or qs2.query.distinct:
            res = qs1.distinct() & qs2.distinct()
        else:
            res = qs1 & qs2

        res._transform_fns = list(set(
            getattr(qs1, '_transform_fns', [])
            + getattr(qs2, '_transform_fns', [])))

        if not (qs1.query.standard_ordering and qs2.query.standard_ordering):
            res.query.standard_ordering = False

        select_related = [qs1.query.select_related, qs2.query.select_related]
        if False in select_related:
            # We are not interested in the default value
            select_related.remove(False)

        if len(select_related) == 1:
            res.query.select_related = select_related[0]
        elif len(select_related) == 2:
            if True in select_related:
                # Prefer explicit select_related to generic select_related()
                select_related.remove(True)

            if len(select_related) > 0:
                # If we have two explicit select_related calls, take any
                res.query.select_related = select_related[0]
            else:
                res = res.select_related()

        res._prefetch_related_lookups = list(
            set(qs1._prefetch_related_lookups)
            | set(qs2._prefetch_related_lookups))

        return res

    while tail:
        head = _merge(head, tail[0])
        tail = tail[1:]
    return head


_KWARG_RE = re.compile("(?:([-\w]+)=)?(.+)")


def parse_args_and_kwargs(parser, bits):
    """
    Parses template tag arguments and keyword arguments

    Returns a tuple ``args, kwargs``.

    Usage::

        @register.tag
        def custom(parser, token):
            return CustomNode(*parse_args_and_kwargs(parser,
                token.split_contents()[1:]))

        class CustomNode(template.Node):
            def __init__(self, args, kwargs):
                self.args = args
                self.kwargs = kwargs

            def render(self, context):
                args, kwargs = resolve_args_and_kwargs(context, self.args,
                    self.kwargs)
                return self._render(context, *args, **kwargs):

            def _render(self, context, ...):
                # The real workhorse
    """
    args = []
    kwargs = {}

    for bit in bits:
        match = _KWARG_RE.match(bit)
        key, value = match.groups()
        value = parser.compile_filter(value)
        if key:
            kwargs[str(key)] = value
        else:
            args.append(value)

    return args, kwargs


def resolve_args_and_kwargs(context, args, kwargs):
    """
    Resolves arguments and keyword arguments parsed by
    ``parse_args_and_kwargs`` using the passed context instance

    See ``parse_args_and_kwargs`` for usage instructions.
    """
    return [v.resolve(context) for v in args], dict((k, v.resolve(context))
        for k, v in kwargs.items())


def changed_regions(regions, fields):
    """
    Returns a subset of regions which have to be updated when fields have
    been edited. To be used together with the ``{% regions %}`` template
    tag.

    Usage::

        regions = {}
        render(request, 'detail.html', {
            'object': instance,
            'regions': regions,
            })
        return HttpResponse(
            json.dumps(changed_regions(regions, ['emails', 'phones'])),
            content_type='application/json')
    """
    dependencies = regions.get('_dependencies', {})
    to_update = set(itertools.chain(*[
        dependencies.get(field, []) for field in fields]))

    return dict((key, value) for key, value in regions.items()
        if key in to_update)


def tryreverse(*args, **kwargs):
    """
    Calls ``django.core.urlresolvers.reverse``, and returns ``None`` on
    failure instead of raising an exception.
    """
    try:
        return reverse(*args, **kwargs)
    except NoReverseMatch:
        return None


def substitute_with(to_delete, instance):
    """
    Substitute the first argument with the second in all relations,
    and delete the first argument afterwards.
    """
    assert to_delete.__class__ == instance.__class__
    assert to_delete.pk != instance.pk

    for related_object in to_delete._meta.get_all_related_objects():
        queryset = related_object.model._base_manager.complex_filter({
            related_object.field.name: to_delete.pk,
        })

        queryset.update(**{related_object.field.name: instance.pk})
    to_delete.delete()


def app_model_label(model):
    """
    Stop those deprecation warnings
    """
    try:
        return model._meta.app_label, model._meta.model_name
    except AttributeError:  # Django <1.6
        return model._meta.app_label, model._meta.module_name



"""
This module beefs up the default full text search field to be a little
bit more versatile. It allows specifying patterns such as ``is:unread``
or ``!important`` which are extracted from the query string and returned
as standalone values allowing the implementation of a search syntax
known from f.e. Google Mail.

Quick rules always consist of two parts: A regular expression pulling
values out of the query string and a mapper which maps the values from
the regex to something else which may be directly usable by forms.

Usage example::

    QUICK_RULES = [
        (re.compile(r'!!'), quick.static(important=True)),
        (re.compile(r'@(?P<username>\w+)'),
            quick.model_mapper(User.objects.all(), 'assigned_to')),
        (re.compile(r'\^\+(?P<due>\d+)'),
            lambda v: {'due': date.today() + timedelta(days=int(v['due']))}),
        (re.compile(r'=(?P<estimated_hours>[\d\.]+)h'),
            quick.identity()),
        ]

    data, rest = quick.parse_quickadd(
        request.POST.get('quick', u''),
        QUICK_RULES)

    data['notes'] = u' '.join(rest) # Everything which could not be parsed
                                    # is added to the ``notes`` field.
    form = TicketForm(data)

.. note::

   The mappers always get the regex matches ``dict`` and return a
   ``dict``.
"""

from datetime import date, timedelta

from django.utils import dateformat
from django.utils.datastructures import MultiValueDict
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _


def parse_quickadd(quick, regexes):
    """
    The main workhorse. Named ``parse_quickadd`` for historic reasons,
    can be used not only for adding but for searching etc. too. In fact,
    :class:`towel.forms.SearchForm` supports quick rules out of the box
    when they are specified in ``quick_rules``.
    """

    data = {}
    rest = []

    while quick:
        for regexp, extract in regexes:
            match = regexp.match(quick)
            if match:
                for key, value in extract(match.groupdict()).items():
                    data.setdefault(key, []).append(value)

                quick = quick[len(match.group(0)):].strip()
                break

        else:
            splitted = quick.split(' ', 1)
            if len(splitted) < 2:
                rest.append(quick)
                break

            rest.append(splitted[0])
            quick = splitted[1]

    return MultiValueDict(data), rest


def identity():
    """
    Identity mapper. Returns the values from the regular expression
    directly.
    """
    return (lambda value: value)


def model_mapper(queryset, attribute):
    """
    The regular expression needs to return a dict which is directly passed
    to ``queryset.get()``. As a speciality, this mapper returns both the
    primary key of the instance under the ``attribute`` name, and the instance
    itself as ``attribute_``.
    """
    def _fn(values):
        try:
            instance = queryset.get(**values)
            return {
                attribute: instance.pk,
                attribute + '_': instance,
            }
        except (queryset.model.DoesNotExist, KeyError, TypeError, ValueError):
            return {}
    return _fn


def static(**kwargs):
    """
    Return a predefined ``dict`` when the given regex matches.
    """
    return (lambda values: kwargs)


def model_choices_mapper(data, attribute):
    """
    Needs a ``value`` provided by the regular expression and returns
    the corresponding ``key`` value.

    Example::

        class Ticket(models.Model):
            VISIBILITY_CHOICES = (
                ('public', _('public')),
                ('private', _('private')),
                )
            visibility = models.CharField(choices=VISIBILITY_CHOICES)

        QUICK_RULES = [
            (re.compile(r'~(?P<value>[^\s]+)'), quick.model_choices_mapper(
                Ticket.VISIBILITY_CHOICES, 'visibility')),
            ]
    """
    def _fn(values):
        reverse = dict((force_text(value), key) for key, value in data)
        try:
            return {attribute: reverse[values['value']]}
        except KeyError:
            return {}
    return _fn


def due_mapper(attribute):
    """
    Understands ``Today``, ``Tomorrow``, the following five localized
    week day names or (partial) dates such as ``20.12.`` and ``01.03.2012``.
    """
    def _fn(values):
        today = date.today()
        due = values['due']

        days = [(dateformat.format(d, 'l'), d) for d in [
            (today + timedelta(days=d)) for d in range(2, 7)]]
        days.append((_('Today'), today))
        days.append((_('Tomorrow'), today + timedelta(days=1)))
        days = dict((k.lower(), value) for k, value in days)

        if due.lower() in days:
            return {attribute: days[due.lower()]}

        day = [today.year, today.month, today.day]
        try:
            for i, n in enumerate(due.split('.')):
                day[2 - i] = int(n, 10)
        except (IndexError, TypeError, ValueError):
            pass

        try:
            return {attribute: date(*day)}
        except (TypeError, ValueError):
            pass

        return {}
    return _fn


def bool_mapper(attribute):
    """
    Maps ``yes``, ``1`` and ``on`` to ``True`` and ``no``, ``0``
    and ``off`` to ``False``.
    """
    def _fn(values):
        if values['bool'].lower() in ('yes', '1', 'on', 'true'):
            return {attribute: True}
        elif values['bool'].lower() in ('no', '0', 'off', 'false'):
            return {attribute: False}
        return {}
    return _fn



import json

from django import forms
from django.db import models
from django.db.models import ObjectDoesNotExist
from django.forms.util import flatatt
from django.http import HttpResponse, QueryDict
from django.utils import six
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from towel import quick


class BatchForm(forms.Form):
    """
    This form class can be used to provide batch editing functionality
    in list views, similar to Django's admin actions.

    You have to implement your batch processing in the ``_context()``
    method. This method only receives one parameter, a queryset which
    is already filtered according to the selected items on the list view.
    Additionally, the current request is available as an attribute of the
    form instance, ``self.request``.

    The method ``process(self)`` may have the following return values:

    - A ``dict`` instance: Will be merged into the template context.
    - A ``HttpResponse`` instance: Will be returned directly to the client.
    - An iterable: The handler assumes successful processing of all objects
      contained in the iterable.
    - Nothing: Nothing happens.

    Usage example::

        class AddressBatchForm(BatchForm):
            subject = forms.CharField()
            body = forms.TextField()

            def process(self):
                # Form validation has already been taken care of
                subject = self.cleaned_data.get('subject')
                body = self.cleaned_data.get('body')

                if not (subject and body):
                    return {}

                sent = 0
                for item in self.batch_queryset:
                    send_mail(subject, body, settings.DEFAULT_SENDER,
                        [item.email])
                    sent += 1
                if sent:
                    messages.success(self.request, 'Sent %s emails.' % sent)

                return self.batch_queryset

        def addresses(request):
            queryset = Address.objects.all()
            batch_form = AddressBatchForm(request, queryset)
            ctx = {'addresses': queryset}

            if batch_form.should_process():
                result = form.process()
                if isinstance(result, HttpResponse):
                    return result
                elif isinstance(result, dict):
                    ctx.update(result)
                elif hasattr(result, '__iter__'):
                    messages.success(request,
                        _('Processed the following items: %s') % (
                            u', '.join(force_text(item) for item in result)))

                return HttpResponseRedirect('.')

            return render(request, 'addresses.html', ctx)

    Template code::

        {% load towel_batch_tags %}
        <form method="post" action=".">
            <ul>
            {% for address in addresses %}
                <li>
                {% batch_checkbox address.id batch_form %}
                {{ address }}
                </li>
            {% endfor %}
            </ul>

            {# Required! Otherwise, ``BatchForm.process`` does nothing. #}
            <input type="hidden" name="batchform" value="1" />

            <table>
                {{ batch_form }}
            </table>
            <button type="submit">Send mail to selected</button>
        </form>
    """

    _process = False
    ids = []

    def __init__(self, request, queryset, *args, **kwargs):
        kwargs.setdefault('prefix', 'batch')

        self.request = request
        self.queryset = queryset

        if request.method == 'POST' and 'batchform' in request.POST:
            self._process = True
            super(BatchForm, self).__init__(request.POST, request.FILES,
                *args, **kwargs)
        else:
            super(BatchForm, self).__init__(*args, **kwargs)

    def clean(self):
        """
        Cleans the batch form fields and checks whether at least one item
        had been selected.
        """
        data = super(BatchForm, self).clean()

        post_data = self.request.POST
        self.ids = [
            pk for pk in self.queryset.values_list('id', flat=True)
            if post_data.get('batch_%s' % pk)]

        if not self.ids:
            raise forms.ValidationError(_('No items selected'))

        return data

    def should_process(self):
        """
        Returns true when the submitted form was the batch form, and the
        batch form is valid.
        """
        return self._process and self.is_valid()

    @cached_property
    def batch_queryset(self):
        """
        Returns the queryset containing only items that have been selected
        for batch processing.
        """
        return self.queryset.filter(id__in=self.ids)

    def process(self):  # pragma: no cover
        """
        Actually processes the batch form submission. Override this with
        your own behavior.

        Batch forms may return the following types here (they are handled
        by ``ModelView.handle_batch_form``:

        - A ``HttpResponse``:
          Will be returned directly to the user.
        - An iterable:
          A success message will be generated containing all items in the
          iterable.
        """
        if hasattr(self, '_context'):
            import warnings
            warnings.warn(
                'The batch form \'%s.%s\' is still using \'_context\'.'
                ' Switch to using the new \'process\' method now!' % (
                    self.__class__.__module__,
                    self.__class__.__name__,
                ),
                DeprecationWarning,
            )

            ctx = self._context(self.batch_queryset)
            if 'response' in ctx:
                return ctx['response']
            return ctx

        raise NotImplementedError(
            'BatchForm.process has no default implementation.')


class SearchForm(forms.Form):
    """
    Supports persistence of searches (stores search in the session). Requires
    not only the GET parameters but the request object itself to work
    correctly.

    Usage example::

        class AddressManager(SearchManager):
            search_fields = ('first_name', 'last_name', 'address', 'email',
                'city', 'zip_code', 'created_by__email')

        class Address(models.Model):
            ...

            objects = AddressManager()

        class AddressSearchForm(SearchForm):
            orderings = {
                '': ('last_name', 'first_name'), # Default
                'dob': 'dob', # Sort by date of birth
                'random': lambda queryset: queryset.order_by('?'),
                }
            is_person = forms.NullBooleanField()

        def addresses(request):
            search_form = AddressSearchForm(request.GET, request=request)
            queryset = search_form.queryset(Address)
            ctx = {
                'addresses': queryset,
                'search_form': search_form,
                }
            return render(request, 'addresses.html', ctx)

    .. warning::

        All fields in the form need to have ``required=False`` set. Otherwise,
        form validation would already fail on the first visit on the list
        page (which would kind of defeat the purpose of a search form).

    Template code::

        <form method="get" action=".">
            <input type="hidden" name="s" value="1"> <!-- SearchForm search -->
            <table>
                {{ search_form }}
            </table>
            <button type="submit">Search</button>
        </form>

        {% for address in addresses %}
            ...
        {% endfor %}
    """

    #: Fields which are always excluded from automatic filtering
    #: in ``apply_filters``
    always_exclude = ('s', 'query', 'o')

    #: Default field values - used if not overridden by the user
    default = {}

    #: Ordering specification
    orderings = {}

    #: Quick rules, a list of (regex, mapper) tuples
    quick_rules = []

    #: Search form active?
    s = forms.CharField(required=False, widget=forms.HiddenInput(),
                        initial='1')

    #: Current ordering
    o = forms.CharField(required=False, widget=forms.HiddenInput())

    #: Full text search query
    query = forms.CharField(required=False, label=_('Query'),
        widget=forms.TextInput(attrs={'placeholder': _('Query')}))

    def __init__(self, data, *args, **kwargs):
        # Are the results filtered in any way?
        self.filtered = True
        # Is a persisted search active?
        self.persistency = False

        request = kwargs.pop('request')
        self.original_data = data
        super(SearchForm, self).__init__(self.prepare_data(data, request),
            *args, **kwargs)
        self.persist(request)
        self.post_init(request)

    def prepare_data(self, data, request):
        """
        Fill in default values from ``default`` if they aren't provided by
        the user.
        """

        if not self.default:
            return data

        data = data.copy()
        for key, value in self.default.items():
            if key not in data:
                if hasattr(value, '__call__'):
                    value = value(request)

                if hasattr(value, '__iter__'):
                    data.setlist(key, value)
                else:
                    data[key] = value
        return data

    def post_init(self, request):
        """
        Hook for customizations.
        """

        pass

    def persist(self, request):
        """
        Persist the search in the session, or load saved search if user
        isn't searching right now.
        """

        session_key = 'sf_%s.%s' % (
            self.__class__.__module__,
            self.__class__.__name__)

        if 'clear' in request.GET or 'n' in request.GET:
            if session_key in request.session:
                del request.session[session_key]

        if self.original_data and (
                set(self.original_data.keys()) & set(self.fields.keys())):
            data = self.data.copy()
            if 's' in data:
                del data['s']
                request.session[session_key] = data.urlencode()

        elif request.method == 'GET' and 's' not in request.GET:
            # try to get saved search from session
            if session_key in request.session:
                self.data = QueryDict(request.session[session_key])
                self.persistency = True

            else:
                self.filtered = False

        elif request.method == 'POST' and 's' not in request.POST:
            # It wasn't the search form which was POSTed, hopefully :-)
            self.filtered = False

    def searching(self):
        """
        Returns ``searching`` for use as CSS class if results are filtered
        by this search form in any way.
        """

        if self.persistency or self.safe_cleaned_data.get('s'):
            return 'searching'
        return ''

    @property
    def safe_cleaned_data(self):
        """
        Safely return a dictionary of values, even if search form isn't valid.
        """

        self.is_valid()
        try:
            return self.cleaned_data.copy()
        except AttributeError:
            return {}

    def fields_iterator(self):
        """
        Yield all additional search fields.
        """

        skip = ('query', 's', 'o')

        for field in self:
            if field.name not in skip:
                yield field

    def apply_filters(self, queryset, data, exclude=()):
        """
        Automatically apply filters

        Uses form field names for ``filter()`` argument construction.
        """

        exclude = list(exclude) + list(self.always_exclude)

        for field in self.fields.keys():
            if field in exclude:
                continue

            value = data.get(field)
            if (value and hasattr(value, '__iter__')
                    and not isinstance(value, six.string_types)):
                queryset = queryset.filter(**{'%s__in' % field: value})
            elif value or value is False:
                queryset = queryset.filter(**{field: value})

        if self.quick_rules:
            quick_only = set(data.keys()) - set(self.fields.keys())
            for field in quick_only:
                if field in exclude:
                    continue

                if field.endswith('_') and (
                        field[:-1] in quick_only or field[:-1] in self.fields):
                    # Either ``quick.model_mapper`` wanted to trick us and
                    # added the model instance, too, or the quick mechanism
                    # filled an existing form field which means that the
                    # attribute has already been handled above.
                    continue

                value = data.get(field)
                if value is not None:
                    queryset = queryset.filter(**{field: value})

        return queryset

    def apply_ordering(self, queryset, ordering=None):
        """
        Applies ordering if the value in ``o`` matches a key in
        ``self.orderings``. The ordering may also be reversed,
        in which case the ``o`` value should be prefixed with
        a minus sign.
        """
        if ordering is None:
            return queryset

        if ordering and ordering[0] == '-':
            order_by, desc = ordering[1:], True
        else:
            order_by, desc = ordering, False

        if order_by not in self.orderings:
            # Ignore ordering request if unknown
            return queryset

        order_by = self.orderings[order_by]

        if hasattr(order_by, '__call__'):
            queryset = order_by(queryset)
        elif hasattr(order_by, '__iter__'):
            queryset = queryset.order_by(*order_by)
        else:
            queryset = queryset.order_by(order_by)

        if desc:
            return queryset.reverse()
        return queryset

    def query_data(self):
        """
        Return a fulltext query and structured data which can be converted into
        simple filter() calls
        """

        if not hasattr(self, '_query_data_cache'):
            data = self.safe_cleaned_data

            if self.quick_rules:
                data, query = quick.parse_quickadd(data.get('query'),
                    self.quick_rules)
                query = u' '.join(query)

                # Data in form fields overrides any quick specifications
                for k, v in self.safe_cleaned_data.items():
                    if v is not None:
                        data[k] = v
            else:
                query = data.get('query')

            self._query_data_cache = query, data
        return self._query_data_cache

    def queryset(self, model):
        """
        Return the result of the search
        """

        query, data = self.query_data()
        queryset = model.objects.search(query)
        queryset = self.apply_filters(queryset, data)
        return self.apply_ordering(queryset, data.get('o'))


class WarningsForm(forms.BaseForm):
    """
    Form subclass which allows implementing validation warnings

    In contrast to Django's ``ValidationError``, these warnings may
    be ignored by checking a checkbox.

    The warnings support consists of the following methods and properties:

    * ``WarningsForm.add_warning(<warning>)``: Adds a new warning message
    * ``WarningsForm.warnings``: A list of warnings or an empty list if there
      are none.
    * ``WarningsForm.is_valid()``: Overridden ``Form.is_valid()``
      implementation which returns ``False`` for otherwise valid forms with
      warnings, if those warnings have not been explicitly ignored (by checking
      a checkbox or by passing ``ignore_warnings=True`` to ``is_valid()``.
    * An additional form field named ``ignore_warnings`` is available - this
      field should only be displayed if ``WarningsForm.warnings`` is non-emtpy.
    """
    def __init__(self, *args, **kwargs):
        super(WarningsForm, self).__init__(*args, **kwargs)

        self.fields['ignore_warnings'] = forms.BooleanField(
            label=_('Ignore warnings'), required=False)
        self.warnings = []

    def add_warning(self, warning):
        """
        Adds a new warning, should be called while cleaning the data
        """
        self.warnings.append(warning)

    def is_valid(self, ignore_warnings=False):
        """
        ``is_valid()`` override which returns ``False`` for forms with warnings
        if these warnings haven't been explicitly ignored
        """
        if not super(WarningsForm, self).is_valid():
            return False

        if self.warnings and not (ignore_warnings or
                self.cleaned_data.get('ignore_warnings')):
            return False

        return True


class StrippedInputMixin(object):
    def value_from_datadict(self, data, files, name):
        value = data.get(name, None)
        if isinstance(value, six.string_types):
            return value.strip()
        return value


class StrippedTextInput(StrippedInputMixin, forms.TextInput):
    """
    ``TextInput`` form widget subclass returning stripped contents only
    """
    pass


class StrippedTextarea(StrippedInputMixin, forms.Textarea):
    """
    ``Textarea`` form widget subclass returning stripped contents only
    """
    pass


def towel_formfield_callback(field, **kwargs):
    """
    Use this callback as ``formfield_callback`` if you want to use stripped
    text inputs and textareas automatically without manually specifying the
    widgets. Adds a ``dateinput`` class to date and datetime fields too.
    """

    if isinstance(field, models.CharField) and not field.choices:
        kwargs['widget'] = StrippedTextInput()
    elif isinstance(field, models.TextField):
        kwargs['widget'] = StrippedTextarea()
    elif isinstance(field, models.DateTimeField):
        kwargs['widget'] = forms.DateTimeInput(attrs={'class': 'dateinput'})
    elif isinstance(field, models.DateField):
        kwargs['widget'] = forms.DateInput(attrs={'class': 'dateinput'})

    return field.formfield(**kwargs)


#: Backwards compatibility, provide the function under the old name
#: ``stripped_formfield_callback`` too.
def stripped_formfield_callback(field, **kwargs):
    import warnings
    warnings.warn(
        'stripped_formfield_callback has been renamed to'
        ' towel_formfield_callback, please start using the new name.',
        DeprecationWarning, stacklevel=2)
    return towel_formfield_callback(field, **kwargs)


def autocompletion_response(queryset, limit=10):
    """
    Helper which returns a ``HttpResponse`` list of instances in a format
    suitable for consumption by jQuery UI Autocomplete, respectively
    ``towel.forms.ModelAutocompleteWidget``.
    """
    return HttpResponse(
        json.dumps([
            {
                'label': force_text(instance),
                'value': instance.pk,
            } for instance in queryset[:limit]
        ]),
        content_type='application/json',
    )


class ModelAutocompleteWidget(forms.TextInput):
    """
    Model autocompletion widget using jQuery UI Autocomplete

    Supports both querysets and JSON-returning AJAX handlers as data
    sources. Use as follows::

        class MyForm(forms.ModelForm):
            customer = forms.ModelChoiceField(Customer.objects.all(),
                widget=ModelAutocompleteWidget(url='/customers/search_ajax/'),
                )
            type = forms.ModelChoiceField(Type.objects.all(),
                widget=ModelAutocompleteWidget(queryset=Type.objects.all()),
                )

    You need to make sure that the jQuery UI files are loaded correctly
    yourself.
    """

    def __init__(self, attrs=None, url=None, queryset=None):
        assert (url is None) != (queryset is None), ('Provide either url'
            ' or queryset')

        self.url = url
        self.queryset = queryset
        super(ModelAutocompleteWidget, self).__init__(attrs)

    def render(self, name, value, attrs=None, choices=()):
        if value is None:
            value = ''
        final_attrs = self.build_attrs(attrs, type='hidden', name=name)
        if value != '':
            # Only add the 'value' attribute if a value is non-empty.
            final_attrs['value'] = force_text(self._format_value(value))

        hidden = u'<input%s />' % flatatt(final_attrs)

        final_attrs['type'] = 'text'
        final_attrs['id'] += '_ac'
        del final_attrs['name']

        try:
            instance = self.choices.queryset.get(pk=value)
            final_attrs['value'] = force_text(instance)
        except (ObjectDoesNotExist, ValueError, TypeError):
            final_attrs['value'] = u''

        if self.is_required:
            ac = u'<input%s />' % flatatt(final_attrs)
        else:
            final_attrs.setdefault('class', '')
            final_attrs['class'] += ' ac_nullable'

            ac = (
                u' <a href="#" id="%(id)s_cl" class="ac_clear">'
                ' %(text)s</a>' % {
                    'id': final_attrs['id'][:-3],
                    'text': _('clear'),
                }
            ) + (u'<input%s />' % flatatt(final_attrs))

        js = u'''<script type="text/javascript">
$(function() {
    $('#%(id)s_ac').autocomplete({
        source: %(source)s,
        focus: function(event, ui) {
            $('#%(id)s_ac').val(ui.item.label);
            return false;
        },
        select: function(event, ui) {
            $('#%(id)s').val(ui.item.value).trigger('change');
            $('#%(id)s_ac').val(ui.item.label);
            return false;
        }
    }).bind('focus', function() {
        this.select();
    }).bind('blur', function() {
        if (!this.value)
            $('#%(id)s').val('');
    });
    $('#%(id)s_cl').click(function(){
        $('#%(id)s, #%(id)s_ac').val('');
        return false;
    });
});
</script>
''' % {'id': attrs.get('id', name), 'name': name, 'source': self._source()}

        return mark_safe(hidden + ac + js)

    def _source(self):
        if self.url:
            if hasattr(self.url, '__call__'):
                return u'\'%s\'' % self.url()
            return u'\'%s\'' % self.url
        else:
            data = json.dumps([
                {
                    'label': force_text(o),
                    'value': o.id,
                } for o in self.queryset.all()
            ])

            return u'''function (request, response) {
    var data = %(data)s, ret = [], term = request.term.toLowerCase();
    for (var i=0; i<data.length; ++i) {
        if (data[i].label.toLowerCase().indexOf(term) != -1)
            ret.push(data[i]);
    }
    response(ret);
}
''' % {'data': data}


class InvalidEntry(object):
    pk = None


class MultipleAutocompletionWidget(forms.TextInput):
    """
    You should probably use harvest chosen instead.
    """
    def __init__(self, attrs=None, queryset=None):
        self.queryset = queryset
        super(MultipleAutocompletionWidget, self).__init__(attrs)

    def _possible(self):
        return dict((force_text(o).lower(), o) for o in self.queryset._clone())

    def render(self, name, value, attrs=None, choices=()):
        if value is None:
            value = []
        final_attrs = self.build_attrs(attrs, name=name, type='text')

        if value:
            value = u', '.join(force_text(o) for o in
                self.queryset.filter(id__in=value))

        js = u'''<script type="text/javascript">
$(function() {
    function split( val ) {
        return val.split( /,\s*/ );
    }
    function extractLast( term ) {
        return split( term ).pop();
    }

    $( "#%(id)s" )
        // don't navigate away from the field on tab when selecting an item
        .bind( "keydown", function( event ) {
            if ( event.keyCode === $.ui.keyCode.TAB &&
                    $( this ).data( "autocomplete" ).menu.active ) {
                event.preventDefault();
            }
        })
        .autocomplete({
            source: %(source)s,
            search: function() {
                // custom minLength
                var term = extractLast( this.value );
                if ( term.length < 2 ) {
                    return false;
                }
            },
            focus: function() {
                // prevent value inserted on focus
                return false;
            },
            select: function( event, ui ) {
                var terms = split( this.value );
                // remove the current input
                terms.pop();
                // add the selected item
                terms.push( ui.item.value );
                // add placeholder to get the comma-and-space at the end
                terms.push( "" );
                this.value = terms.join( ", " );
                return false;
            }
    });
});
</script>
''' % {
            'id': final_attrs.get('id', name),
            'name': name,
            'source': self._source(),
        }

        return mark_safe(u'<textarea%s>%s</textarea>' % (flatatt(final_attrs),
            value) + js)

    def value_from_datadict(self, data, files, name):
        value = data.get(name, None)
        if not value:
            return []

        possible = self._possible()
        values = [s for s in
            [s.strip() for s in value.lower().split(',')] if s]
        return list(set(possible.get(s, InvalidEntry).pk for s in values))

    def _source(self):
        return u'''function(request, response) {
    response($.ui.autocomplete.filter(%(data)s, extractLast(request.term)));
    }''' % {
            'data': json.dumps(
                [force_text(o) for o in self.queryset._clone()]),
        }



# Straight import from https://github.com/simonw/django-queryset-transform

"""
django_queryset_transform
=========================

Allows you to register a transforming map function with a Django QuerySet
that will be executed only when the QuerySet itself has been evaluated.

This allows you to build optimisations like "fetch all tags for these 10 rows"
while still benefiting from Django's lazy QuerySet evaluation.

For example::

    def lookup_tags(item_qs):
        item_pks = [item.pk for item in item_qs]
        m2mfield = Item._meta.get_field_by_name('tags')[0]
        tags_for_item = Tag.objects.filter(
            item__in = item_pks
        ).extra(select = {
            'item_id': '%s.%s' % (
                m2mfield.m2m_db_table(), m2mfield.m2m_column_name()
            )
        })
        tag_dict = {}
        for tag in tags_for_item:
            tag_dict.setdefault(tag.item_id, []).append(tag)
        for item in item_qs:
            item.fetched_tags = tag_dict.get(item.pk, [])

    qs = Item.objects.filter(name__contains = 'e').transform(lookup_tags)

    for item in qs:
        print item, item.fetched_tags

Prints::

    Winter comes to Ogglesbrook [<sledging>, <snow>, <winter>, <skating>]
    Summer now [<skating>, <sunny>]

But only executes two SQL queries - one to fetch the items, and one to fetch
ALL of the tags for those items.

Since the transformer function can transform an evaluated QuerySet, it
doesn't need to make extra database calls at all - it should work for things
like looking up additional data from a cache.multi_get() as well.

Originally inspired by http://github.com/lilspikey/django-batch-select/



LICENSE
=======

Copyright (c) 2010, Simon Willison.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of Django nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from django.db import models


class TransformQuerySet(models.query.QuerySet):
    def __init__(self, *args, **kwargs):
        super(TransformQuerySet, self).__init__(*args, **kwargs)
        self._transform_fns = []

    def _clone(self, klass=None, setup=False, **kw):
        c = super(TransformQuerySet, self)._clone(klass, setup, **kw)
        c._transform_fns = self._transform_fns[:]
        return c

    def transform(self, *fn):
        c = self._clone()
        c._transform_fns.extend(fn)
        return c

    def iterator(self):
        result_iter = super(TransformQuerySet, self).iterator()
        if self._transform_fns:
            results = list(result_iter)
            for fn in self._transform_fns:
                fn(results)
            return iter(results)
        return result_iter


class TransformManager(models.Manager):
    def get_query_set(self):
        return TransformQuerySet(self.model, using=self._db)

    def transform(self, *fn):
        return self.get_query_set().transform(*fn)
